package core

import (
	"git.akyoto.dev/cli/q/src/build/ast"
	"git.akyoto.dev/cli/q/src/build/config"
	"git.akyoto.dev/cli/q/src/build/errors"
	"git.akyoto.dev/cli/q/src/build/expression"
	"git.akyoto.dev/cli/q/src/build/token"
)

// CompileDefinition compiles a variable definition.
func (f *Function) CompileDefinition(node *ast.Define) error {
	name := node.Name.Text()

	if f.identifierExists(name) {
		return errors.New(&errors.VariableAlreadyExists{Name: name}, f.File, node.Name.Position)
	}

	uses := token.Count(f.Body, token.Identifier, name) - 1

	if uses == 0 {
		return errors.New(&errors.UnusedVariable{Name: name}, f.File, node.Name.Position)
	}

	return f.storeVariableInRegister(name, node.Value, uses)
}

func (f *Function) AddVariable(variable *Variable) {
	if config.Comments {
		f.Comment("%s = %s (%d uses)", variable.Name, variable.Register, variable.Alive)
	}

	scope := f.Scope()
	variable.Scope = scope

	scope.variables[variable.Name] = variable
	scope.Use(variable.Register)
}

func (f *Function) useVariable(variable *Variable) {
	for i, scope := range f.scopes {
		if scope.inLoop && variable.Scope != scope {
			continue
		}

		local := scope.variables[variable.Name]

		if local == nil {
			continue
		}

		local.Alive--

		if local.Alive < 0 {
			panic("incorrect number of variable use calls")
		}

		if local.Alive == 0 {
			if config.Comments {
				f.Comment("%s (%s) died in scope %d", local.Name, local.Register, i)
			}

			scope.Free(local.Register)
		} else if config.Comments {
			f.Comment("%s (%s) used in scope %d", local.Name, local.Register, i)
		}
	}
}

// identifierExists returns true if the identifier has been defined.
func (f *Function) identifierExists(name string) bool {
	variable := f.Variable(name)

	if variable != nil {
		return true
	}

	_, exists := f.Functions[name]
	return exists
}

func (f *Function) storeVariableInRegister(name string, value *expression.Expression, uses int) error {
	reg := f.Scope().MustFindFree(f.cpu.General)
	f.Scope().Reserve(reg)
	err := f.ExpressionToRegister(value, reg)

	f.AddVariable(&Variable{
		Name:     name,
		Register: reg,
		Alive:    uses,
	})

	return err
}
